/*
 * IM installer for Win32
 *
 *    Main routine.
 *
 *                    Copyright (C) 1997-1999  Shuichi Kitaguchi
 */

#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <process.h>

#include "../common/env.h"
#include "../common/common.h"

#include "resource.h"


#define  IMINST_VERSION   "0.6.3"
#define  IMINST_DEFAULT   "im.ini"
#define  LOGFILE_NAME     "im.log"


HINSTANCE hInst;
HANDLE    hConsole;

char  szCurrentPath[MAX_PATH];
char  szIniFile[MAX_PATH];

char  szBinDir[MAX_PATH];
char  szDB_Type[32];		/* ok? */
char  szSrcs[BUFLEN];
char  szLibs[BUFLEN];
char  szBins[BUFLEN];
char  szStartup[BUFLEN];
char  szIMDir[MAX_PATH];
char  szPrefix[MAX_PATH];
char  szLibDir[MAX_PATH];

char szMailAddress[256];
char szPopserver[256];
char szSmtpserver[256];
char szUserName[256];
char szDomainName[256];

BOOL  fFull    = FALSE;
BOOL  fCDROM   = FALSE;
BOOL  fDebug   = FALSE;

BOOL FindFiles( LPCSTR lpPrefix, LPCSTR lpPattern, LPSTR lpBuf, DWORD dwLen )
{
  WIN32_FIND_DATA wfd;
  HANDLE hFind;

  *lpBuf = '\0';
  hFind = FindFirstFile( lpPattern, &wfd );
  if ( hFind == INVALID_HANDLE_VALUE ) return FALSE;
  do {
    if ( dwLen <= strlen(lpBuf)+strlen(wfd.cFileName)+1 ) return FALSE;
    if ( lpPrefix )
      strcat( lpBuf, lpPrefix );
    strcat( lpBuf, wfd.cFileName );
    strcat( lpBuf, " " );
  } while ( FindNextFile( hFind, &wfd ) );
  return TRUE;
}

BOOL ReadIniFile( LPCSTR szFileName )
{
  HMODULE  hModule;
  char     *pt;
  char     szBuf[256];
  char     szPBuf[256];

  /* Ini file name */

  hModule = GetModuleHandle(NULL);
  GetModuleFileName(hModule,szCurrentPath,sizeof(szCurrentPath));
  if ( (pt = strrchr(szCurrentPath,'\\')) )
    *pt = (char)0;
  if ( szFileName == NULL ){
    sprintf(szIniFile,"%s\\%s",szCurrentPath,IMINST_DEFAULT);
  } else {
    strcpy(szIniFile,szFileName);
  }

  sprintf(szPBuf,"INI file = [%s]",szIniFile);
  OutputDebugLog(szPBuf);

  if ( ! ReadPerlEnvironments(szIniFile) ) return ( FALSE );
  if ( fDebug ) PrintPerlEnvironments();

  GetPrivateProfileString("Make","Full","no",
			  szBuf,sizeof(szBuf),
			  szIniFile);
  if ( ! strcmp(szBuf,"yes") ) fFull = TRUE;

  GetPrivateProfileString("Make","PREFIX","Default",
			  szPrefix,sizeof(szPrefix),
			  szIniFile);
  if ( ! strcmp(szPrefix,"Default") )
    strcpy(szPrefix,szPerlDir);

  GetPrivateProfileString("Make","LIBDIR","Default",
			  szLibDir,sizeof(szLibDir),
			  szIniFile);
  if ( ! strcmp(szLibDir,"Default") )
    strcpy(szLibDir,szPerlLib);
  sprintf(szIMDir,"%s\\IM",szLibDir);

  GetPrivateProfileString("Make","BINDIR","Default",
			  szBinDir,sizeof(szBinDir),
			  szIniFile);
  if ( ! strcmp(szBinDir,"Default") )
    strcpy(szBinDir,szPerlBin);
  sprintf(szPBuf,"BINDIR=[%s]",szBinDir);
  OutputDebugLog(szPBuf);

  GetPrivateProfileString("Make","DB_TYPE","Default",
			  szDB_Type,sizeof(szDB_Type),
			  szIniFile);
  GetPrivateProfileString("Make","SRCS","Default",
			  szSrcs,sizeof(szSrcs),
			  szIniFile);
  if ( ! strcmp(szSrcs, "Default") ){
    FindFiles( NULL, "im*.in", szSrcs, sizeof(szSrcs) );
  }
  GetPrivateProfileString("Make","LIBS","Default",
			  szLibs,sizeof(szLibs),
			  szIniFile);
  if ( ! strcmp(szLibs, "Default") ){
    FindFiles( "IM/", "IM/*.pm", szLibs, sizeof(szLibs) );
    strcat( szLibs, "IM/Config.pm" ); /* exception */
  }
  GetPrivateProfileString("Make","BINS","Default",
			  szBins,sizeof(szBins),
			  szIniFile);
  GetPrivateProfileString("Make","STARTUP","Default",
			  szStartup,sizeof(szStartup),
			  szIniFile);
  if ( fDebug ){
    sprintf(szPBuf,"DB_Type=[%s]",szDB_Type);
    OutputLog(szPBuf);
    sprintf(szPBuf,"Srcs=[%s]",szSrcs);
    OutputLog(szPBuf);
    sprintf(szPBuf,"Libs=[%s]",szLibs);
    OutputLog(szPBuf);
    sprintf(szPBuf,"Bins=[%s]",szBins);
    OutputLog(szPBuf);
    sprintf(szPBuf,"Startup=[%s]",szStartup);
    OutputLog(szPBuf);
  }

  return ( TRUE );
}


BOOL MakeIMDirectories( VOID )
{
  char szPBuf[256];
  BOOL fError=FALSE;

  /* Make directories(if not exist) */
  if ( ! CheckDirectory( szIMDir ) ){
    sprintf(szPBuf,"[%s] is not exist. creating...",szIMDir);
    if ( ! CreateDirectory( szIMDir, NULL ) ){
      strcat(szPBuf,"fail.");
      fError = TRUE;
    } else {
      strcat(szPBuf,"ok.");
    }
    OutputLog(szPBuf);
  }
#if 0
  if ( ! CheckDirectory( szBinDir ) ){
    sprintf(szPBuf,"[%s] is not exist. creating...",szBinDir);
    if ( ! CreateDirectory( szBinDir, NULL ) ){
      strcat(szPBuf,"fail.");
      fError = TRUE;
    } else {
      strcat(szPBuf,"ok.");
    }
    OutputLog(szPBuf);
  }
#endif
  if ( fError ) return ( FALSE );
  return ( TRUE );
}


BOOL ConvertFile( LPCSTR szSrcFile, LPCSTR szDstFile )
{
  char buf[BUFLEN];
  FILE *sfp,*dfp;
  char szBuf[256];
  char *pt;

  strcpy( szBuf, szPerlLib );
  pt = strrchr( szBuf, '/' );
  if ( pt ) *pt = '\0';

  if ( (sfp = fopen(szSrcFile,"r")) == NULL )
    return ( FALSE );
  if ( (dfp = fopen(szDstFile,"w")) == NULL ){
    fclose(sfp);
    return ( FALSE );
  }
  while ( fgets( buf, sizeof(buf), sfp) ){
    ReplaceString( buf, "@prefix@", szPrefix );
    ReplaceString( buf, "@exec_prefix@", szBuf );
    ReplaceString( buf, "@libdir@", szLibDir );

    ReplaceString( buf, "@im_db_type@",      szDB_Type );
    ReplaceString( buf, "@im_file_attr@",    "Fcntl::O_RDWR()|Fcntl::O_CREAT()|Fcntl::O_EXCL()|Fcntl::O_BINARY()" );
    ReplaceString( buf, "@im_src_siteperl@", "" ); /* xxx */
    ReplaceString( buf, "@im_my_siteperl@",  "" ); /* xxx */
    ReplaceString( buf, "@im_path_ssh@",     "ssh" ); /* xxx */
    ReplaceString( buf, "@im_path_perl@",    szPerl );
    ReplaceString( buf, "@im_use_cl@",       "no" );
    ReplaceString( buf, "@im_fsync_no@",     "0" ); /* xxx */
    fputs(buf,dfp);
  }
  fclose(dfp);
  fclose(sfp);

  return ( TRUE );
}


BOOL ConvertSources( VOID )
{
  char *token;
  char seps[] = " \t";
  char szBuf[BUFLEN],szFileName[MAX_PATH];
  char *pt;
  BOOL fError=FALSE;
  char szPBuf[256];


  if ( ! MakeIMDirectories() ) return ( FALSE );

  /* Commands */
  strcpy(szBuf,szSrcs);
  token = strtok( szBuf, seps );
  while ( token ){
    if ( ! fCDROM ){
      strcpy(szFileName,token);
      pt = strrchr(szFileName,'.');
      *pt = (char)0;		/* remove ".in" */
    } else {
      sprintf(szFileName,"%s\\%s",szBinDir,token);
      pt = strrchr(szFileName,'.');
      *pt = (char)0;		/* remove ".in" */
    }
    sprintf(szPBuf,"Converting [%s] to [%s] ...",token,szFileName);
    if ( ! ConvertFile(token,szFileName) ){
      strcat(szPBuf,"fail.");
      OutputLog(szPBuf);
      return ( FALSE );
    } else {
      strcat(szPBuf,"ok.");
      OutputDebugLog(szPBuf);
    }
    token = strtok( NULL, seps );
  } /* while ( token ){ */

  /* SiteConfig */
  if ( ! fCDROM ){
    strcpy( szFileName, "cnf.im/SiteConfig" );
  } else {
    sprintf( szFileName, "%s/SiteConfig", szIMDir );
  }
  sprintf( szPBuf, "Converting [cnf.im/SiteConfig.in] to [%s] ...", szFileName );
  if ( ! ConvertFile("cnf.im/SiteConfig.in", szFileName ) ){
    strcat(szPBuf,"fail.");
    OutputLog(szPBuf);
    return ( FALSE );
  } else {
    strcat(szPBuf,"ok.");
    OutputDebugLog(szPBuf);
  }

  /* Config.pm */
  if ( ! fCDROM ){
    strcpy( szFileName, "IM/Config.pm");
  } else {
    sprintf( szFileName, "%s/Config.pm", szIMDir );
  }
  sprintf( szPBuf, "Converting [IM/Config.pm.in] to [%s] ...", szFileName);
  if ( ! ConvertFile("IM/Config.pm.in",szFileName) ){
    strcat(szPBuf,"fail.");
    OutputLog(szPBuf);
    return ( FALSE );
  } else {
    strcat(szPBuf,"ok.");
    OutputDebugLog(szPBuf);
  }

  return ( TRUE );
}


HDDEDATA CALLBACK DdeCallback( UINT     uType,
			       UINT     uFmt,
			       HCONV    hconv,
			       HSZ      hsz1,
			       HSZ      hsz2,
			       HDDEDATA hdata,
			       DWORD    dwData1,
			       DWORD    dwData2 )
{
  return NULL;
}


/* copy files */
BOOL InstallIM( VOID )
{
  char *token;
  char seps[] = " \t";
  char szBuf[MAX_PATH],szSrcFile[MAX_PATH];
  char *pt;
  BOOL fError = FALSE;
  char szPBuf[256];

  if ( ! MakeIMDirectories() ) return ( FALSE );

  if ( ! fCDROM ){
    /* Commands */
    token = strtok( szSrcs, seps );
    while ( token ){
      pt = strrchr(token,'.');
      *pt = (char)0;
      sprintf(szBuf,"%s\\%s",szBinDir,token);
      sprintf(szPBuf,"Copying [%s] to [%s] ...",token,szBuf);
      if ( ! CopyFile(token,szBuf,FALSE) ){
	strcat(szPBuf,"fail.");
	fError = TRUE;
      } else {
	strcat(szPBuf,"ok.");
      }
      OutputDebugLog(szPBuf);  

      token = strtok( NULL, seps );
    } /* while ( token ){ */
    if ( fError ){
      OutputLog("Error: cannot copy command files.\n");
      return ( FALSE );
    }
  } /* if ( ! fCDROM ){ */

  /* Libraries */
  token = strtok( szLibs, seps );
  while ( token ){
    sprintf(szBuf,"%s/%s",szLibDir,token);
    sprintf(szPBuf,"Copying [%s] to [%s] ...",token,szBuf);
    if ( ! CopyFile(token,szBuf,FALSE) ){
      strcat(szPBuf,"fail.");
      OutputLog(szPBuf);
      return ( FALSE );
    } else {
      strcat(szPBuf,"ok.");
      OutputDebugLog(szPBuf);
    }
    token = strtok( NULL, seps );
  } /* while ( token ){ */

  if ( ! fCDROM ){
    /* cnf.im/SiteConfig */
    sprintf(szBuf,"%s\\SiteConfig",szIMDir);
    sprintf(szPBuf,"Copying [cnf.im/SiteConfig] to [%s] ...",szBuf);
    if ( ! CopyFile("cnf.im/SiteConfig",szBuf,FALSE) ){
      strcat(szPBuf,"fail.");
      OutputLog(szPBuf);
      return ( FALSE );
    } else {
      strcat(szPBuf,"ok.");
      OutputDebugLog(szPBuf);
    }
  } /* if ( ! fCDROM ){ */

  /* Bins */
  if ( fFull && strcmp(szBins,"Default") ){
    token = strtok( szBins, seps );
    while ( token ){
      sprintf(szBuf,"%s\\%s",szBinDir,token);
      sprintf(szSrcFile,"Win32\\%s",token);
      sprintf(szPBuf,"Copying [%s] to [%s] ...",szSrcFile,szBuf);
      if ( ! CopyFile(szSrcFile,szBuf,FALSE) ){
	strcat(szPBuf,"fail.");
	OutputLog(szPBuf);
	return ( FALSE );
      } else {
	strcat(szPBuf,"ok.");
	OutputDebugLog(szPBuf);
      }
      token = strtok( NULL, seps );
    }
  }

  return ( TRUE );
}


BOOL CALLBACK SetupIMDlgProc( HWND   hDlg,
			      UINT   uiMsg,
			      WPARAM wParam,
			      LPARAM lParam )
{
  BOOL fRet = TRUE;
  char *pt;

  switch ( uiMsg ){
  case WM_INITDIALOG:
    ShowWindow( hDlg, SW_SHOW );
    UpdateWindow( hDlg );
    break;

  case WM_COMMAND:
    switch ( GET_WM_COMMAND_ID( wParam, lParam ) ){
    case IDOK:
      GetDlgItemText( hDlg, IDC_MAILADDRESS, szMailAddress, sizeof(szMailAddress) );
      if ( ! strcmp( szMailAddress, "" ) ){
	BilErrorMessageBox( "Mail Address is empty.",
			    "Mail Address ̗󔒂ł" );
	break;
      }
      pt = strchr( szMailAddress, '@' );
      if ( pt != NULL ){
	*pt = '\0';
	if ( *(pt+1) != '\0' ){
	  strcpy(szUserName,szMailAddress);
	  strcpy(szDomainName, pt+1 );
	  *pt = '@';
	} /* if ( *(pt+1) != '\0' ){ */
      } else {
	BilErrorMessageBox( "Invalid mail address and retry.",
			    "[AhXsłBēx͂ĉ" );
	SetDlgItemText( hDlg, IDC_MAILADDRESS, "" );
	fRet = FALSE;
	break;
      } /* if ( pt != NULL ){ */
      
      GetDlgItemText( hDlg, IDC_SMTPSERVER, szSmtpserver, sizeof(szSmtpserver) );
      if ( ! strcmp( szSmtpserver, "" ) ){
	BilErrorMessageBox( "Smtpserver is empty.",
			    "Smtpserver ̗󔒂ł" );
	break;
      }
      GetDlgItemText( hDlg, IDC_POPSERVER, szPopserver, sizeof(szPopserver) );
      if ( ! strcmp( szPopserver, "" ) ){
	BilErrorMessageBox( "Popserver is empty.",
			    "Popserver ̗󔒂ł" );
	break;
      }
      EndDialog( hDlg, 0 );
      break;
    case IDCANCEL:
      fRet = FALSE;
      EndDialog( hDlg, -1 );
      break;
    }
  default:
    fRet = FALSE;
    break;
  } /* switch ( uiMsg ){ */
  
  return ( fRet );
}


/* setup for IM */
BOOL SetupIM( VOID )
{
  BOOL fError=FALSE;
  char szBuf[MAX_PATH];
  FILE *sfp,*dfp;
  BOOL fOldConfFile=FALSE;
  char szPBuf[256];

  /* check IM installation */
  sprintf(szBuf,"%s\\imget",szBinDir);
  if ( ! CheckFile( szBuf ) ){
    BilErrorMessageBox( "IM is not installed.",
			"IM CXg[Ă܂B" );
    return ( FALSE );
  }

  if ( DialogBoxParam( hInst,
		       MAKEINTRESOURCE(IDD_SETUP),
		       NULL,
		       SetupIMDlgProc,
		       0 )
       != 0 ) return ( FALSE );

  /* make ~/.im */
  if ( ! CheckDirectory( szConfPath ) ){
    sprintf(szPBuf,"[%s] is not exit. creating...",szConfPath);
    if ( ! CreateDirectory(szConfPath,NULL) ){
      strcat(szPBuf,"fail.");
      fError = TRUE;
    } else {
      strcat(szPBuf,"ok.");
    }
    OutputLog(szPBuf);
  }
  if ( fError ) return ( FALSE );
  
  /* copy Config to ~/.im/ */
  if ( CheckFile(szConfFile) ){
    fOldConfFile=TRUE;
    OutputDebugLog("old Conf file exists. copy to Conf.new file.");
    strcat(szConfFile,".new");
  }
  if ( (sfp = fopen("dot.im/Config","r")) == NULL )
    return ( FALSE );
  if ( (dfp = fopen(szConfFile,"w")) == NULL ){
    fclose(sfp);
    return ( FALSE );
  }
  while ( fgets( szBuf, sizeof(szBuf), sfp) ){
    ReplaceString( szBuf, "#MailDir=Mail", "MailDir=Mail" );
    ReplaceString( szBuf, "#InboxFolder=+inbox", "InboxFolder=+inbox" );
    ReplaceString( szBuf, "#DraftFolder=+draft", "DraftFolder=+draft" );
    ReplaceString( szBuf, "#TrashFolder=+trash", "TrashFolder=+trash" );
    ReplaceString( szBuf, "#AddrBookFile=Addrbook", "AddrBookFile=Addrbook" );

    if ( strstr( szBuf, "#User=" ) )
      sprintf( szBuf, "User=%s\n", szUserName );
    if ( strstr( szBuf, "Address=" ) )
      sprintf( szBuf, "Address=%s\n", szMailAddress );
    if ( strstr( szBuf, "FromDomain=_domain_of_your_mail_address_" ) )
      sprintf( szBuf, "FromDomain=%s\n", szDomainName );
    if ( strstr( szBuf, "ToDomain=" ) )
      sprintf( szBuf, "ToDomain=%s\n", szDomainName );
    if ( strstr( szBuf, "#Imget.Src=local\t" ) )
      sprintf( szBuf, "Imget.Src=POP/POP:%s@%s\n", szUserName, szPopserver );
    if ( strstr( szBuf, "#Smtpservers=localhost" ) )
      sprintf( szBuf, "Smtpservers=%s\n", szSmtpserver );
    
    fputs(szBuf,dfp);
  }
  fclose(dfp);
  fclose(sfp);
  
  /* make directory Mail/News */
  sprintf(szBuf,"%s\\Mail",szHomePath);
  if ( ! CheckDirectory( szBuf ) ){
    sprintf(szPBuf,"[%s] is not exist. creating...",szBuf);
    if ( ! CreateDirectory(szBuf,NULL) ){
      strcat(szPBuf,"fail.");
      fError = TRUE;
    } else {
      strcat(szPBuf,"ok.");
    }
    OutputLog(szPBuf);
  }
  if ( fError ) return ( FALSE );

  if ( fOldConfFile ){
    BilInfoMessageBox( "Installer creates new configuration file.\nPlease copy ~/.im/Config.new to ~/.im/Config.",
		       "Vݒt@C쐬܂\nKvɉ ~/.im/Config.new  ~/.im/Config ɃRs[ĉ");
  } else {
    BilInfoMessageBox( "Installer creates configuration file.\nPlease check ~/.im/Config file.",
		       "ݒt@C쐬܂\n~/.im/Config t@C`FbNĉ");
  }
  
  return ( TRUE );
}



/* add programs to startup folder */
BOOL InstallStartup( VOID )
{
  char  seps[] = " \t";
  char  *token;
  char  szBuf[MAX_PATH*2+100],szPrgBuf[MAX_PATH];
  DWORD dwDde=0;
  HCONV HConversation;
  HSZ   hszPM;

  if ( ! strcmp( szStartup, "Default" ) ) return ( FALSE );
  if ( ! fFull ) return ( FALSE );

  DdeInitialize (&dwDde, (PFNCALLBACK)DdeCallback, APPCMD_CLIENTONLY, 0);
  hszPM = DdeCreateStringHandle (dwDde, "PROGMAN", CP_WINANSI);

  if ( HConversation = DdeConnect (dwDde, hszPM, hszPM, NULL) ){
    token = strtok( szStartup, seps );
    while ( token ){
      sprintf(szPrgBuf,"%s\\%s",szBinDir,token);
      
      strcpy(szBuf,"[CreateGroup (Mew)]");
      DdeClientTransaction (szBuf, strlen(szBuf)+1, HConversation,
			    (HSZ)NULL, CF_TEXT, XTYP_EXECUTE, 30000, NULL);

      strcpy(szBuf,"[ReplaceItem (Password Agent)]");
      DdeClientTransaction (szBuf, strlen(szBuf)+1, HConversation,
			    (HSZ)NULL, CF_TEXT, XTYP_EXECUTE, 30000, NULL);

      sprintf(szBuf, "[AddItem (%s,Password Agent,%s)]",
	      szPrgBuf,szPrgBuf);
      DdeClientTransaction (szBuf, strlen(szBuf)+1, HConversation,
			    (HSZ)NULL, CF_TEXT, XTYP_EXECUTE, 30000, NULL);
      
      token = strtok( NULL, seps );
    } /* while ( token ){ */
    DdeDisconnect (HConversation);
  } /* if (HConversation = DdeConnect (dwDde, hszPM, hszPM, NULL)){ */
  
  DdeFreeStringHandle (dwDde, hszPM );
  DdeUninitialize (dwDde);

  return ( TRUE );
}

VOID ShowUsage( VOID )
{
  char buf[65536];

  _snprintf(buf,sizeof(buf),
	    "IM installer for Win32  Version %s\n"
	    "          Copyright (C) 1997-1999  Shuichi Kitaguchi\n"
	    "Usage : im.exe [options]\n"
	    "Option:\n"
	    "        -s     setup mode.\n"
	    "        -d     debug mode.  show detail message.\n"
	    "        -v     verbose mode.  write message to [%s]\n"
	    "        -h     show this message.\n"
	    "        -q     quiet mode.  do not show any messages.\n"
	    "        -cd    installation only (not compile)\n",
	    IMINST_VERSION,LOGFILE_NAME
	    );
  MessageBox( NULL, buf, "Usage", MB_OK|MB_SETFOREGROUND|MB_ICONINFORMATION);
}

/* Main */
int WINAPI WinMain( HINSTANCE hInstance,
		    HINSTANCE hPrevInstance,
		    LPSTR     lpCmdLine,
		    int       nCmdShow )
{
  BOOL   fVerbose = FALSE;
  BOOL   fLogTime = FALSE;
  BOOL   fSetup   = FALSE;
  BOOL   fQuiet   = FALSE;
  HANDLE hMutex;
  char   szPBuf[256];
  char   szEPBuf[256];
  char   szJPBuf[256];
  char   seps[] = " \t";
  char   *token;

  hInst = hInstance;

  token = strtok( lpCmdLine, seps );
  while ( token ){
    if ( ! strcmp(token,"-cd") )
      fCDROM = TRUE;		/* CD-ROM mode */
    if ( ! strcmp(token,"-s") )
      fSetup = TRUE;		/* setup mode */
    if ( ! strcmp(token,"-d") )
      fDebug = TRUE;		/* debug mode */
    if ( ! strcmp(token,"-q") )
      fQuiet = TRUE;		/* quiet mode */
    if ( ! strcmp(token,"-v") )
      fVerbose = TRUE;		/* verbose mode */
    if ( ! strcmp(token,"-h") ){
      ShowUsage();
      return 0;
    }
    token = strtok( NULL, seps );
  } /* while ( token ){ */

  SetLogMode( LOGFILE_NAME, fDebug, fLogTime, fVerbose, 0, fQuiet );

  /* parent process's console? xxx */
  AllocConsole();
  hConsole = CreateConsoleScreenBuffer( GENERIC_READ|GENERIC_WRITE,
					0,
					NULL,
					CONSOLE_TEXTMODE_BUFFER,
					NULL );
  if ( hConsole == INVALID_HANDLE_VALUE ) goto Error;
  if ( ! SetStdHandle( STD_OUTPUT_HANDLE, hConsole ) ){
    OutputLog("SetStdhandle fails.");
    goto Error;
  }
  if ( ! SetConsoleActiveScreenBuffer( hConsole ) ){
    OutputLog("SetConsoleActiveScreenBuffer fails.");
    goto Error;
  }

  sprintf(szPBuf,"IM installer for Win32  Version %s",IMINST_VERSION);
  OutputLog(szPBuf);
  OutputLog("          Copyright (C) 1997-1999  Shuichi Kitaguchi");

  if ( ! GetEnvironments() ) goto Error;
  if ( ! ReadEnvironments() ) goto Error;
  PrintEnvironments();

  /* PwAgent existent check */
  hMutex = CreateMutex( NULL,TRUE,PWAGENT_MUTEX_NAME );
  if ( hMutex != NULL ){
    if ( GetLastError() == ERROR_ALREADY_EXISTS ){
      BilErrorMessageBox( "Please exit PwAgent and try to install.",
			  "PwAgent IĂCXg[sĉB" );
      goto Error;
    }
  } else {
    OutputLog("Error: cannot create Mutex.");
    goto Error;
  }
  ReleaseMutex( hMutex );
  CloseHandle( hMutex );

  /* ini file check */
  if ( ! CheckFile(IMINST_DEFAULT) ){
    sprintf(szEPBuf,"[%s] is not found!",IMINST_DEFAULT);
    sprintf(szJPBuf,"[%s] t܂!",IMINST_DEFAULT);
    BilErrorMessageBox( szEPBuf, szJPBuf );
    goto Error;
  }

  if ( ! CheckFile("configure.in") ){
    BilErrorMessageBox( "Archive Extraction Error!\nCheck Archiver Program(support long filename?).",
			"A[JCủ𓀂ɎsĂ܂\n𓀃vOOt@Cl[T|[gĂ邩AmFĉ");
    goto Error;
  }

  if ( langId == LANGID_JAPANESE ){
    OutputLog(">>INI t@Cǂݍł܂...");
  } else {
    OutputLog(">>Read INI File...");
  }
  if ( ! ReadIniFile(NULL) ) goto Error;
  if ( langId == LANGID_JAPANESE ){
    strcpy(szPBuf,">>>Perl  [");
    strcat(szPBuf,szPerl);
    strcat(szPBuf,"] ɂ܂B");
  } else {
    strcpy(szPBuf,">>>Perl exists in [");
    strcat(szPBuf,szPerl);
    strcat(szPBuf,"]");
  }
  OutputLog(szPBuf);

  if ( fSetup ){
    if ( langId == LANGID_JAPANESE ){
      OutputLog(">>ݒt@C쐬Ă܂...");
    } else {
      OutputLog(">>Setup Configuration File...");
    }
    if ( ! SetupIM() ) goto Error;
    return ( 0 );
  } /* if ( fSetup ){ */

  /* ConvertSources/ConvertFile/InstallIM */
  ConvertPathSeparator(szPrefix);
  ConvertPathSeparator(szLibDir);
  ConvertPathSeparator(szPerl);
  ConvertPathSeparator(szPerlLib);
  if ( langId == LANGID_JAPANESE ){
    OutputLog(">>\[Xt@CϊĂ܂...");
  } else {
    OutputLog(">>Convert sources...");
  }
  if ( ! ConvertSources() ) goto Error;
  
  if ( langId == LANGID_JAPANESE ){
    OutputLog(">>CXg[Ă܂...");
  } else {
    OutputLog(">>Install All Files...");
  }
  if ( ! InstallIM() ) goto Error;
  
  if ( fFull && strcmp(szStartup,"Default") ){
    if ( langId == LANGID_JAPANESE ){
      OutputLog(">>vOX^[gj[ɓo^Ă܂...");
    } else {
      OutputLog(">>Add Programs to Startup folder...");
    }
    if ( ! InstallStartup() ) goto Error;
  }
  BilInfoMessageBox(
		    "IM installation Complete!\nIf you use IM for the first time, execute \"im.exe -s\".",
		    "IM ̃CXg[I܂\nIM ߂Ďgꍇ́A\"im.exe -s\" sĉ");

  return ( 0 );
  
 Error:
  BilErrorMessageBox( "IM installation is NOT complete!",
		      "IM ̃CXg[ɏI܂ł" );
  return ( -1 );
}
