/*
 * Password Agent for Win32
 *
 *    Main Source.
 *
 *                    Copyright (C) 1997-1999  Shuichi Kitaguchi
 */
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <stdio.h>
#include <process.h>

#include "pwagent.h"
#include "../common/env.h"
#include "../common/common.h"

#include "resource.h"


ImPassword     *impass=NULL;	/* IM passwords */
BOOL           fSaveImPasswords=FALSE; /* if save passwords? */

HANDLE         hWaitEvent=NULL;
unsigned long  hMainThread=0;
HINSTANCE      hInst=NULL;	/* instance handle */
SYSTEMTIME     st;
HANDLE         hMutex=NULL;
UINT           uiThreadId=0;

SOCKET         msgsock;
unsigned short port_num;	/* port number */

HMENU          hMenu=NULL, hSubMenu=NULL; /* menu handle */
HICON          hIcon=NULL;	/* icon handle */
NOTIFYICONDATA nIcon;		/* notify icon data */
int            iCurPage=0;	/* current page */

char tc_item[PWNUM_MAX][8] = { /* tab control menu string */
  "0", "1", "2", "3", "4", "5", "6", "7"
};



/*****************************************************************
 *                     Configuration                             *
 *****************************************************************/


unsigned short GetPortNum( void )
{
  char szBuf[256];

  ReadConfig("PwAgentPort",szBuf,sizeof(szBuf));
  return ( ( szBuf[0] == 0 ) ? PWAGENT_PORT : (unsigned short)atoi(szBuf) );
}



/*****************************************************************
 *                   Menu Management                             *
 *****************************************************************/


VOID EnableAllMenu( VOID )
{
  EnableMenuItem( hSubMenu, IDM_ABOUT, MF_ENABLED );
  EnableMenuItem( hSubMenu, IDM_PASSWORD, MF_ENABLED );
  EnableMenuItem( hSubMenu, IDM_EXIT, MF_ENABLED );
}

VOID DisableAllMenu( VOID )
{
  EnableMenuItem( hSubMenu, IDM_ABOUT, MF_DISABLED|MF_GRAYED );
  EnableMenuItem( hSubMenu, IDM_PASSWORD, MF_DISABLED|MF_GRAYED );
  EnableMenuItem( hSubMenu, IDM_EXIT, MF_DISABLED|MF_GRAYED );
}


/*****************************************************************
 *                  Password Management                          *
 *****************************************************************/


void InitImPassword( ImPassword *imp )
{
  memset(imp,0,sizeof(ImPassword)); /* all clear */
}

void ClearAllImPasswords( void )
{
  int i;

  if ( impass )
    for ( i=0; i<PWNUM_MAX; i++ )
      InitImPassword( &impass[i] );
}

void PrintImPassword( ImPassword *imp )
{
  char  szPBuf[256];

  sprintf(szPBuf,"prot     = [%s]",imp->prot);
  OutputDebugLog(szPBuf);
  sprintf(szPBuf,"auth     = [%s]",imp->auth);
  OutputDebugLog(szPBuf);
  sprintf(szPBuf,"server   = [%s]",imp->server);
  OutputDebugLog(szPBuf);
  sprintf(szPBuf,"user     = [%s]",imp->user);
  OutputDebugLog(szPBuf);
  sprintf(szPBuf,"password = [%s]",imp->password);
  OutputDebugLog(szPBuf);
}

BOOL AllocateImPassword( int iNum )
{
  impass = (ImPassword*)malloc(sizeof(ImPassword)*PWNUM_MAX);
  if ( impass == NULL ) return ( FALSE );
  ClearAllImPasswords();
  
  return ( TRUE );
}

BOOL FreeImPasswords( void )
{
  if ( impass ){
    free(impass);
    return ( TRUE );
  } else {
    return ( FALSE );
  }  
}

BOOL CompareImPassword( ImPassword *a, ImPassword *b )
{
  if ( _stricmp(a->prot,b->prot) ) return ( FALSE );
  if ( _stricmp(a->auth,b->auth) ) return ( FALSE );
  if ( _stricmp(a->server,b->server) ) return ( FALSE );
  if ( _stricmp(a->user,b->user) ) return ( FALSE );

  return ( TRUE );
}

ImPassword* SearchImPassword( ImPassword *imp )
{
  int i;

  if ( impass == NULL ) return ( NULL );

  for ( i=0; i<PWNUM_MAX; i++ )
    if ( CompareImPassword(&impass[i],imp) )
      return ( &impass[i] );

  return ( NULL );
}

ImPassword* SearchEmptyImPassword( void )
{
  int i;

  if ( impass == NULL ) return ( NULL );

  for ( i=0; i<PWNUM_MAX; i++ )
    if ( ! _stricmp( impass[i].server, "" ) )
      return ( &impass[i] );

  return ( NULL );
}

BOOL SaveImPassword( ImPassword *imp )
{
  ImPassword *im;

  if ( impass == NULL ) return ( FALSE );

  if ( (im = SearchImPassword(imp)) == NULL ){
    if ( (im = SearchEmptyImPassword()) == NULL ){
      impass[0] = *imp;		/* no space! */
    } else {
      *im = *imp;		/* space found */
    }
  } else {
    /* update data */
    strcpy(im->user,imp->user);
    strcpy(im->password,imp->password);
  }
  return ( TRUE );
}

void LoadImPassword( ImPassword *imp )
{
  ImPassword *im;

  if ( impass == NULL ) return;

  if ( (im = SearchImPassword(imp)) == NULL ){
    imp->password[0] = 0;	/* not found */
  } else {
    strcpy(imp->password,im->password);	/* found */
  }
}

BOOL SaveImPasswordsToReg( BOOL fSavePass )
{
  int    i;
  LONG   lRet;
  HKEY   hKey;
  char   szSubKey[MAX_PATH];

  if ( impass == NULL ) return ( FALSE );

  /* create registry keys */
  lRet = RegOpenKeyEx( HKEY_CURRENT_USER,
		       PWAG_SUBKEY,
		       0,
		       KEY_READ|KEY_WRITE,
		       &hKey );
  if ( lRet != ERROR_SUCCESS ){
    OutputDebugLog("SaveImPasswordsToReg: Error: cannot open registry.");
    return ( FALSE );
  }

  lRet = RegSetValueEx( hKey,
			PWAG_SUBKEY_SAVEPASS,
			0,
			REG_BINARY,
			(const BYTE*)&fSaveImPasswords,
			sizeof(BYTE) );
  if ( lRet != ERROR_SUCCESS ){
    OutputDebugLog("WinMain: Error: cannot read registry value.");
    RegCloseKey( hKey );
    return ( INIT_ERROR_INIT );
  } /* if ( lRet != ERROR_SUCCESS ){ */

  if ( fSavePass ){
    lRet = RegSetValueEx( hKey,
			  PWAG_SUBKEY_SYSTEMTIME,
			  0,
			  REG_BINARY,
			  (const BYTE*)&st,
			  sizeof(SYSTEMTIME) );
    if ( lRet != ERROR_SUCCESS ){
      OutputDebugLog("WinMain: Error: cannot read registry value.");
      RegCloseKey( hKey );
      return ( INIT_ERROR_INIT );
    } /* if ( lRet != ERROR_SUCCESS ){ */
    
    for ( i=0; i<PWNUM_MAX; i++ ){
      sprintf(szSubKey,"pass%d",i);
      lRet = RegSetValueEx( hKey,
			    szSubKey,
			    0,
			    REG_BINARY,
			    (const BYTE*)&impass[i],
			    sizeof(ImPassword) );
      if ( lRet != ERROR_SUCCESS ){
	OutputDebugLog("SaveImPasswordsToReg: Error: cannot set registry value.");
	RegCloseKey( hKey );
	return ( FALSE );
      }	/* if ( lRet != ERROR_SUCCESS ){ */
    } /*  for ( i=0; i<PWNUM_MAX; i++ ){ */
  } else {
    RegDeleteKey( hKey, PWAG_SUBKEY_SYSTEMTIME );
    for ( i=0; i<PWNUM_MAX; i++ ){
      sprintf(szSubKey,"pass%d",i);
      RegDeleteKey( hKey, szSubKey );
    } /*  for ( i=0; i<PWNUM_MAX; i++ ){ */
  } /* if ( fSavePass ){ */

  RegCloseKey( hKey );

  return ( TRUE );
}

BOOL LoadImPasswordsFromReg( VOID )
{
  int    i;
  LONG   lRet;
  HKEY   hKey;
  DWORD  dwData;
  DWORD  dwType;
  char   szSubKey[MAX_PATH];

  if ( impass == NULL ) return ( FALSE );

  /* create registry keys */
  lRet = RegOpenKeyEx( HKEY_CURRENT_USER,
		       PWAG_SUBKEY,
		       0,
		       KEY_READ|KEY_WRITE,
		       &hKey );
  if ( lRet != ERROR_SUCCESS ){
    OutputDebugLog("SaveImPasswordsToReg: Error: cannot open registry.");
    return ( FALSE );
  }

  dwType = REG_BINARY;
  dwData = sizeof(SYSTEMTIME);
  lRet = RegQueryValueEx( hKey,
			  PWAG_SUBKEY_SYSTEMTIME,
			  0,
			  &dwType,
			  (LPBYTE)&st,
			  &dwData );
  if ( lRet != ERROR_SUCCESS ){
    OutputDebugLog("SaveImPasswordsToReg: Error: cannot query registry value.");
    RegCloseKey( hKey );
    return ( FALSE );
  }

  for ( i=0; i<PWNUM_MAX; i++ ){
    sprintf(szSubKey,"pass%d",i);
    dwType = REG_BINARY;
    dwData = sizeof(ImPassword);
    lRet = RegQueryValueEx( hKey,
			    szSubKey,
			    0,
			    &dwType,
			    (LPBYTE)&impass[i],
			    &dwData );
    if ( lRet != ERROR_SUCCESS ){
      OutputDebugLog("SaveImPasswordsToReg: Error: cannot query registry value.");
      RegCloseKey( hKey );
      return ( FALSE );
    } /* if ( lRet != ERROR_SUCCESS ){ */
  } /*  for ( i=0; i<PWNUM_MAX; i++ ){ */
  RegCloseKey( hKey );
  
  return ( TRUE );
}


/*****************************************************************
 *                  Communication Routines                       *
 *****************************************************************/


UINT WINAPI MainThread( void *dummy )
{
  char                szPBuf[256];
  char                buffer[1024];
  int                 ret,fromlen;
  struct sockaddr_in  local, from;
  SOCKET              listen_socket;
  ImPassword          imp;
  char                *token;
  char                seps[] = "\t\r\n";
  int                 i;

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("MainThread: enter.");
#endif

  listen_socket = socket( AF_INET, SOCK_STREAM, 0 );
  if ( listen_socket == INVALID_SOCKET ){
    sprintf(szPBuf,"MainThread: Error: socket() fail. (%d)",WSAGetLastError());
    OutputDebugLog(szPBuf);
    return ( (UINT)-1 );
  }

  local.sin_family = AF_INET;
  local.sin_addr.s_addr = inet_addr("127.0.0.1");
  local.sin_port = htons(port_num);
  if ( bind( listen_socket, (struct sockaddr*)&local, sizeof(local) ) ==
       SOCKET_ERROR ){
    sprintf(szPBuf,"MainThread: Error: bind() fail. (%d)",WSAGetLastError());
    OutputDebugLog(szPBuf);
    closesocket(listen_socket);
    return ( (UINT)-1 );
  }

  if ( listen( listen_socket, 5 ) == SOCKET_ERROR ){
    sprintf(szPBuf,"MainThread: Error: listen() fail. (%d)",WSAGetLastError());
    OutputDebugLog(szPBuf);
    closesocket(listen_socket);
    return ( (UINT)-1 );
  }

  /* Initialize Complete */
  SetEvent( hWaitEvent );

  while ( 1 ){
    fromlen = sizeof(from);
    msgsock = accept( listen_socket, (struct sockaddr*)&from, &fromlen );
    if ( msgsock == INVALID_SOCKET ){
      sprintf(szPBuf,"MainThread: Error: accept() error. (%d)",WSAGetLastError());
      OutputDebugLog(szPBuf);
      closesocket(listen_socket);
      return ( (UINT)-1 );
    }
    
#ifdef DETAIL_ERROR_MESSAGE
    sprintf(szPBuf,"accepted connection from %s, port %d",
	    inet_ntoa(from.sin_addr),htons(from.sin_port));
    OutputDebugLog(szPBuf);
#endif

    buffer[0] = (st.wMilliseconds>>8)&0xff;
    buffer[1] = (st.wMilliseconds&0xff);
    for ( i=0; i<2; i++ )
      buffer[i] = (buffer[i] < 0x20) ? buffer[i]|0x20 : buffer[i];
    buffer[2] = 0x0a;
    buffer[3] = 0;
    ret = send( msgsock, buffer, strlen(buffer), 0 );
    if ( ret == SOCKET_ERROR ){
      sprintf(szPBuf,"MainThread: Error: send() fail. (%d)",WSAGetLastError());
      OutputDebugLog(szPBuf);
      closesocket(msgsock);
      continue;
    }

    ret = recv( msgsock, buffer, sizeof(buffer), 0 );
    if ( ret == SOCKET_ERROR ){
      sprintf(szPBuf,"MainThread: Error: recv() fail. (%d)",WSAGetLastError());
      OutputDebugLog(szPBuf);
      closesocket(msgsock);
      continue;
    }
    if ( ret == 0 ){
      sprintf(szPBuf,"MailThread: client closed connection");
      OutputDebugLog(szPBuf);
      closesocket(msgsock);
      continue;
    }
    buffer[ret] = 0;
#ifdef _DEBUG
    sprintf(szPBuf,"recv: %d bytes/[%s]",ret,buffer);
    OutputDebugLog(szPBuf);
#endif
    /***** execute commands - start - *****/
    if ( ! strncmp(buffer,"CLEAR",5) ){
      OutputDebugLog("[clear]");
      ClearAllImPasswords();
      strcpy(buffer,"CLEARED\n");
      ret = send( msgsock, buffer, sizeof(buffer), 0 );
      if ( ret == SOCKET_ERROR ){
	sprintf(szPBuf,"MainThread: Error: send() fail. (%d)",WSAGetLastError());
	OutputDebugLog(szPBuf);
      }
    } else if ( ! strncmp(buffer,"SAVE",4) ){
      OutputDebugLog("[save]");
      InitImPassword(&imp);

      token = strtok( buffer, seps );
      token = strtok( NULL, seps );
      if ( sizeof(imp.prot) > strlen(token) )
	strcpy(imp.prot,token);
      token = strtok( NULL, seps );
      if ( sizeof(imp.auth) > strlen(token) )
	strcpy(imp.auth,token);
      token = strtok( NULL, seps );
      if ( sizeof(imp.server) > strlen(token) )
	strcpy(imp.server,token);
      token = strtok( NULL, seps );
      if ( sizeof(imp.user) > strlen(token) )
	strcpy(imp.user,token);
      token = strtok( NULL, seps );
      token = strtok( NULL, seps );
      if ( sizeof(imp.password) > strlen(token) )
	strcpy(imp.password,token);
#ifdef _DEBUG
      PrintImPassword(&imp);
#endif
      if ( (imp.prot[0]!='\0')   &&
	   (imp.auth[0]!='\0')   &&
	   (imp.server[0]!='\0') &&
	   (imp.user[0]!='\0')   &&
	   (imp.password[0]!='\0') ){
	SaveImPassword(&imp);
      }

      strcpy(buffer,"OK\n");
      ret = send( msgsock, buffer, sizeof(buffer), 0 );
      if ( ret == SOCKET_ERROR ){
	sprintf(szPBuf,"MainThread: Error: send() fail. (%d)",WSAGetLastError());
	OutputDebugLog(szPBuf);
      }
    } else if ( ! strncmp(buffer,"LOAD",4) ){
      OutputDebugLog("[load]");
      InitImPassword(&imp);

      token = strtok( buffer, seps );
      token = strtok( NULL, seps );
      if ( sizeof(imp.prot) > strlen(token) )
	strcpy(imp.prot,token);
      token = strtok( NULL, seps );
      if ( sizeof(imp.auth) > strlen(token) )
	strcpy(imp.auth,token);
      token = strtok( NULL, seps );
      if ( sizeof(imp.server) > strlen(token) )
	strcpy(imp.server,token);
      token = strtok( NULL, seps );
      if ( sizeof(imp.user) > strlen(token) )
	strcpy(imp.user,token);

#ifdef _DEBUG
      PrintImPassword(&imp);
#endif
      LoadImPassword(&imp);
#ifdef _DEBUG
      PrintImPassword(&imp);
#endif

      strcpy(buffer,"PASS\t");
      if ( imp.password[0] != 0 ) strcat(buffer,imp.password);
      strcat(buffer,"\n");
      ret = send( msgsock, buffer, sizeof(buffer), 0 );
      if ( ret == SOCKET_ERROR ){
	sprintf(szPBuf,"MainThread: Error: send() fail. (%d)",WSAGetLastError());
	OutputDebugLog(szPBuf);
      }
    } else if ( ! strncmp(buffer,"PING",4) ){
      OutputDebugLog("[ping]");
      strcpy(buffer,"PONG\n");
      ret = send( msgsock, buffer, sizeof(buffer), 0 );
      if ( ret == SOCKET_ERROR ){
	sprintf(szPBuf,"MainThread: Error: send() fail. (%d)",WSAGetLastError());
	OutputDebugLog(szPBuf);
      }
    } else if ( ! strncmp(buffer,"QUIT",4) ){
      OutputDebugLog("[quit]");
      strcpy(buffer,"BYE\n");
      ret = send( msgsock, buffer, sizeof(buffer), 0 );
      if ( ret == SOCKET_ERROR ){
	sprintf(szPBuf,"MainThread: Error: send() fail. (%d)",WSAGetLastError());
	OutputDebugLog(szPBuf);
      }
      closesocket(msgsock);
      break;
    } else {
      OutputDebugLog("[etc]");
      strcpy(buffer,"BYE\n");
      ret = send( msgsock, buffer, sizeof(buffer), 0 );
      if ( ret == SOCKET_ERROR ){
	sprintf(szPBuf,"MainThread: Error: send() fail. (%d)",WSAGetLastError());
	OutputDebugLog(szPBuf);
      }
    }
    /***** execute commands - end - *****/
    closesocket(msgsock);
  } /* while ( 1 ){ */
  closesocket(listen_socket);

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("MainThread: leave.");
#endif

  _endthreadex(0);
  return( 0 );
}



/*****************************************************************
 *                  About Dialog Management                      *
 *****************************************************************/


BOOL CALLBACK AboutDlgProc( HWND   hDlg,
			    UINT   uMsg,
			    WPARAM wParam,
			    LPARAM lParam )
{
  char szBuf[256];
  BOOL fRet=TRUE;
  HWND hTitle;

  switch ( uMsg ){
  case WM_INITDIALOG:
    /* set about dialog text */
    strcpy(szBuf,"Password Agent");
    if ( langId == LANGID_JAPANESE )
      if ( (st.wMilliseconds & 0xf) == 0xf )
	strcpy(szBuf,"˂݂̂͂΂");
    strcat(szBuf," for Win32  Version ");
    strcat(szBuf,PWAGENT_VERSION);
    hTitle = GetDlgItem( hDlg, IDC_PWAGTITLE );
    if ( hTitle ) SetWindowText(hTitle,szBuf);
    
    ShowWindow( hDlg, SW_SHOW );
    UpdateWindow( hDlg );
    break;

  case WM_COMMAND:
    switch ( GET_WM_COMMAND_ID( wParam, lParam ) ){
    case IDOK:
      EndDialog( hDlg, 0 );
      break;
    }
    break;
  default:
    fRet = FALSE;
    break;
  }
  return ( fRet );
}


/*****************************************************************
 *                 Password Dialog Management                    *
 *****************************************************************/


VOID DisplayInfo( HWND hDlg, int num )
{
  char szBuf[32];

  /* set PROTO/AUTH */
  if ( ! _stricmp( impass[num].prot, "" ) ){
    strcpy( szBuf, "" );
  } else {
    sprintf(szBuf,"%s/%s",impass[num].prot,impass[num].auth);
  }
  SetDlgItemText( hDlg, IDC_PROTOCOL, szBuf );

  /* set SERVER NAME */
  SetDlgItemText( hDlg, IDC_SERVERNAME, impass[num].server );

  /* set USER NAME */
  SetDlgItemText( hDlg, IDC_USERNAME, impass[num].user );

  /* set PASSWORD */
  SetDlgItemText( hDlg, IDC_PASSWORD, impass[num].password );

  /* set SAVE PASSWORD FLAG */
  CheckDlgButton( hDlg, IDC_SAVEPASSWORD,
		  (fSaveImPasswords==FALSE) ? BST_UNCHECKED : BST_CHECKED );

  UpdateWindow( hDlg );
}  

  
BOOL CALLBACK PasswordDlgProc( HWND   hDlg,
			       UINT   uMsg,
			       WPARAM wParam,
			       LPARAM lParam )
{
  BOOL     fRet=TRUE;
  int      i;
  TC_ITEM  tci;
  HWND     hTab;
  NMHDR    *lpnmh;
  int      iTab;

  switch ( uMsg ){
  case WM_INITDIALOG:
    hTab = FindWindowEx( hDlg, NULL, NULL, "Tab1" ); /* xxx */
    /* add tab control items */
    for ( i=0; i<PWNUM_MAX; i++ ){
      memset( &tci, 0, sizeof(TC_ITEM) );
      tci.mask       = TCIF_TEXT;
      tci.pszText    = tc_item[i];
      tci.cchTextMax = strlen(tci.pszText)+1;
      TabCtrl_InsertItem( hTab, i, &tci );
    } /* for ( i=0; i<PWNUM_MAX; i++ ){ */

    iCurPage=0;
    DisplayInfo( hDlg, iCurPage );
    break;
    
  case WM_COMMAND:
    switch ( GET_WM_COMMAND_ID( wParam, lParam ) ){
    case IDOK:
      if ( IsDlgButtonChecked( hDlg, IDC_SAVEPASSWORD ) == BST_CHECKED ){
	fSaveImPasswords = TRUE;
	SaveImPasswordsToReg( TRUE );
      } else {
	fSaveImPasswords = FALSE;
	SaveImPasswordsToReg( FALSE );
      }
      EndDialog( hDlg, 0 );
      break;
    case IDCANCEL:
      EndDialog( hDlg, -1 );
      break;
    case IDC_CLEAR:
      InitImPassword( &impass[iCurPage] );
      DisplayInfo( hDlg, iCurPage );
      break;
    case IDC_ALLCLEAR:
      ClearAllImPasswords();
      DisplayInfo( hDlg, iCurPage );
      break;
    } /* switch ( GET_WM_COMMAND_ID( wParam, lParam ) ){ */

  case WM_NOTIFY:
    iTab  = LOWORD(wParam);
    if ( iTab != IDC_TAB ) break;

    lpnmh = (LPNMHDR)lParam;
    iCurPage = TabCtrl_GetCurSel( lpnmh->hwndFrom );
    switch ( lpnmh->code ){
    case TCN_SELCHANGING:
      break;
    case TCN_SELCHANGE:
      if ( IsDlgButtonChecked( hDlg, IDC_SAVEPASSWORD ) == BST_CHECKED ){
	fSaveImPasswords = TRUE;
      } else {
	fSaveImPasswords = FALSE;
      }	
      DisplayInfo( hDlg, iCurPage );
      break;
    } /* switch ( lpnmh->code ){ */
    break;

  default:
    fRet = FALSE;
    break;
  }
  return ( fRet );
}



/*****************************************************************
 *                  Main Window Management                       *
 *****************************************************************/


WORD InitApplication( HWND hWnd )
{
  char   szPBuf[256];
  LONG   lRet;
  HKEY   hKey;
  DWORD  dwDisposition;
  DWORD  dwData;
  DWORD  dwType;

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("InitApplication: enter.");
#endif

  /* init value */
  nIcon.cbSize = 0;

  /* existent check */
  hMutex = CreateMutex( NULL,
			TRUE,
			PWAGENT_MUTEX_NAME );
  if ( hMutex != NULL ){
    if ( GetLastError() == ERROR_ALREADY_EXISTS )
      return( INIT_ERROR_RUNNING );
  } else {
    OutputDebugLog("WinMain: Error: cannot create Mutex.");
    return( INIT_ERROR_INIT );    
  }

  /* allocate password buffer */
  if ( ! AllocateImPassword(0) ){
    OutputDebugLog("InitApplication: Error: cannot allocate password area.");
    return ( INIT_ERROR_INIT );
  }

  /* create registry keys */
  lRet = RegCreateKeyEx( HKEY_CURRENT_USER,
			 PWAG_SUBKEY,
			 0,
			 "",
			 0,
			 KEY_READ|KEY_WRITE,
			 NULL,
			 &hKey,
			 &dwDisposition );
  if ( lRet != ERROR_SUCCESS ){
    OutputDebugLog("WinMain: Error: cannot create registry.");
    return ( INIT_ERROR_INIT );
  }
  dwData = 1;
  dwType = REG_BINARY;
  lRet = RegQueryValueEx( hKey,
			  PWAG_SUBKEY_SAVEPASS,
			  0,
			  &dwType,
			  (LPBYTE)&fSaveImPasswords,
			  &dwData );
  if ( lRet != ERROR_SUCCESS ){
    /* if fail, value is not found. may be... xxx */
    fSaveImPasswords=FALSE;
    lRet = RegSetValueEx( hKey,
			  PWAG_SUBKEY_SAVEPASS,
			  0,
			  REG_BINARY,
			  (const BYTE*)&fSaveImPasswords,
			  sizeof(BYTE) );
    if ( lRet != ERROR_SUCCESS ){
      OutputDebugLog("WinMain: Error: cannot read registry value.");
      RegCloseKey( hKey );
      return ( INIT_ERROR_INIT );
    } /* if ( lRet != ERROR_SUCCESS ){ */
  } /* if ( lRet != ERROR_SUCCESS ){ */
  if ( fSaveImPasswords ){
    LoadImPasswordsFromReg();
  } else {
    /* get system time */
    GetSystemTime(&st);
  }
  RegCloseKey( hKey );

  /* load icon */
  hIcon = LoadIcon( hInst, MAKEINTRESOURCE(IDI_MEW) );

  /* load menu */
  hMenu = LoadMenu( hInst, MAKEINTRESOURCE(IDR_MENU) );
  hSubMenu = GetSubMenu( hMenu, 0 );

  /* set menu */
  nIcon.cbSize           = sizeof(NOTIFYICONDATA);
  nIcon.uID              = IDT_PWAGENT;
  nIcon.hWnd             = hWnd;
  nIcon.uFlags           = NIF_ICON | NIF_MESSAGE | NIF_TIP;
  nIcon.hIcon            = hIcon;
  nIcon.uCallbackMessage = WM_TASKTRAY;
  strcpy(nIcon.szTip,PWAGENT_NAME);
  Shell_NotifyIcon( NIM_ADD, &nIcon );

  /* init socket */
  if ( ! InitSocket() ){
    OutputDebugLog("InitApplication: Error: cannot initialize socket.");
    return ( INIT_ERROR_INIT );
  }

  /* wait event */
  hWaitEvent = CreateEvent (0, TRUE, FALSE, 0);
  if ( ! hWaitEvent ){
    OutputDebugLog("InitApplication: Error: cannot create event.");
    return ( INIT_ERROR_INIT );
  }
  /* main thread */
  hMainThread = _beginthreadex( NULL,
				0,
				MainThread,
				NULL,
				0,
				&uiThreadId );
				
  if ( hMainThread == 0 ){
    sprintf(szPBuf,"InitApplication: Error: cannot create MainThread.(%d)",errno);
    OutputDebugLog(szPBuf);
    return ( INIT_ERROR_INIT );
  }
  /* wait initializing */
  if ( WaitForSingleObject( hWaitEvent, 3000 ) == WAIT_TIMEOUT ){
    OutputDebugLog("InitApplication: Error: cannot initialize MainThread.");
    return ( INIT_ERROR_INIT );
  }
  ResetEvent( hWaitEvent );

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("InitApplication: leave.");
#endif

  return ( INIT_ERROR_NONE );
}


BOOL TermApplication( HWND hWnd )
{
  char szBuf[256];

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("TermApplication: enter.");
#endif

  /* clear registry entry */
  if ( ! fSaveImPasswords ){
    ClearAllImPasswords();
    memset( &st, 0, sizeof(st) );
  }
  SaveImPasswordsToReg( TRUE );

  /* delete menu */
  if ( nIcon.cbSize != 0 )
    Shell_NotifyIcon( NIM_DELETE, &nIcon );
  if ( hMenu != NULL )
    DestroyMenu(hMenu);

  /* stop main thread and waiting */
  if ( hMainThread != 0 ){
    SendCommand("QUIT\n",szBuf);
    if ( WaitForSingleObject( (HANDLE)hMainThread, 3000 ) == WAIT_TIMEOUT )
      OutputDebugLog("TermApplication: Error: WaitForSingleObject timeout.");
    CloseHandle((HANDLE)hMainThread);
  }
  TermSocket();

  /* free memory */
  FreeImPasswords();

  if ( hWaitEvent != NULL )
    CloseHandle( hWaitEvent );
  hWaitEvent = NULL;

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("TermApplication: leave.");
#endif

  return ( TRUE );
}


LRESULT CALLBACK MainWndProc( HWND   hWnd,
			      UINT   uMsg,
			      WPARAM wParam,
			      LPARAM lParam )
{
  POINT pt;
  WORD  wRet;
  UINT  uiRet;

  switch ( uMsg ){
  case WM_CREATE:
    wRet = InitApplication( hWnd );
    if ( wRet == INIT_ERROR_RUNNING ){
      BilErrorMessageBox( "PwAgent is already running.",
			  "PwAgent ͂łɎsł" );
      SendMessage(hWnd, WM_DESTROY, 0, 0);
    } else if ( wRet == INIT_ERROR_INIT ){
      BilErrorMessageBox( "Initialization failed.",
			  "Ɏs܂" );
      SendMessage(hWnd, WM_DESTROY, 0, 0);
    }
    break;
    
  case WM_DRAWITEM:
    break;
    
  case WM_DESTROY:
    TermApplication(hWnd);
    PostQuitMessage(0);
    break;
    
  case WM_COMMAND:
    switch ( GET_WM_COMMAND_ID(wParam, lParam) ){
    case IDOK:
      ShowWindow(hWnd, SW_HIDE);
      break;
    }
    break;
    
  case WM_TASKTRAY:
    if ( lParam == WM_RBUTTONDOWN ){
      GetCursorPos(&pt);
      SetForegroundWindow(hWnd);
      
      uiRet =
	TrackPopupMenu( hSubMenu,
			TPM_BOTTOMALIGN | TPM_LEFTBUTTON | TPM_RIGHTBUTTON |
			TPM_RETURNCMD,
			pt.x, pt.y, 0, hWnd, NULL );
      switch ( uiRet ){
      case IDM_PASSWORD:
	DisableAllMenu();
	DialogBoxParam( hInst,
			MAKEINTRESOURCE(IDD_PASSWORD),
			hWnd,
			PasswordDlgProc,
			(LPARAM)0 );
	EnableAllMenu();
	break;
      case IDM_ABOUT:
	DisableAllMenu();
	DialogBoxParam( hInst,
			MAKEINTRESOURCE(IDD_ABOUT),
			hWnd,
			AboutDlgProc,
			(LPARAM)0 );
	EnableAllMenu();
	break;
      case IDM_EXIT:
	DestroyWindow(hWnd);
	break;
      }	/* switch ( uiRet ){ */
    } /* if ( lParam == WM_RBUTTONDOWN ){ */
    break;

  default:
    return ( DefWindowProc( hWnd, uMsg, wParam, lParam ) );
  }
  
  return ( 0 );
}


/*****************************************************************
 *                      Main Routine                             *
 *****************************************************************/


int WINAPI WinMain( HINSTANCE hInstance,
		    HINSTANCE hPrevInstance,
		    LPSTR     lpCmdLine,
		    int       nCmdShow )
{
  MSG        msg;
  HWND       hWnd;
  WNDCLASSEX wce;
  
  hInst = hInstance;

#ifdef _DEBUG
  SetLogMode(LOGFILE_NAME,TRUE,TRUE,TRUE,0,FALSE);
#else
  SetLogMode(LOGFILE_NAME,FALSE,FALSE,FALSE,0,FALSE);
#endif

  if ( ! ReadEnvironments() ){
    OutputDebugLog("WinMain: Error: cannot read environments.");
    exit( -1 );
  }
  port_num = GetPortNum();

  /* create window */
  wce.hInstance     = hInst;
  wce.lpszClassName = PWAGENT_NAME;
  wce.lpfnWndProc   = MainWndProc;
  wce.style         = 0;
  wce.hIcon = wce.hIconSm = hIcon;
  wce.hCursor       = LoadCursor( NULL, IDI_APPLICATION );
  wce.lpszMenuName  = NULL;
  wce.cbClsExtra = wce.cbWndExtra = 0;
  wce.hbrBackground = (HBRUSH)GetStockObject( LTGRAY_BRUSH );
  wce.cbSize        = sizeof(WNDCLASSEX);
  RegisterClassEx( &wce );

  /* main */
  hWnd = CreateWindowEx( WS_EX_TOOLWINDOW,
			 PWAGENT_NAME,
			 PWAGENT_NAME,
			 WS_OVERLAPPEDWINDOW,
			 CW_USEDEFAULT,
			 CW_USEDEFAULT,
			 100,100,
			 HWND_DESKTOP,
			 NULL,
			 hInst,
			 NULL );
  ShowWindow( hWnd, SW_HIDE );
  UpdateWindow( hWnd );

  /* main loop. */
  while ( GetMessage( &msg, 0, 0, 0 ) ){
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  } /* while ( GetMessage( &msg, 0, 0, 0 ) ){ */

  /* exiting... */
  ReleaseMutex(hMutex);
  CloseHandle(hMutex);

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("WinMain: leave.");
#endif

  return ( msg.wParam );
}
