/*
 * Password Agent for Win32
 *
 *               Socket routines.
 *
 *                    Copyright (C) 1997-1999  Shuichi Kitaguchi
 */

#include <windows.h>
#include <windowsx.h>
#include <stdio.h>


#include "../common/common.h"
#include "pwagent.h"


BOOL InitSocket( void )
{
  WSADATA  wsaData;
  char     szPBuf[256];

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("InitSocket: enter.");
#endif

  if ( WSAStartup( MAKEWORD(1,1), &wsaData ) == SOCKET_ERROR ){
    sprintf(szPBuf,"InitSocket: Error: WSAStartup fail. (%d)",WSAGetLastError());
    OutputDebugLog(szPBuf);
    return ( FALSE );
  }

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("InitSocket: leave.");
#endif

  return ( TRUE );
}


BOOL TermSocket( VOID )
{
#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("TermSocket: enter.");
#endif

  WSACleanup();

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("TermSocket: leave.");
#endif

  return ( TRUE );
}


BOOL SendCommand ( LPCSTR lpszSendCommand,
				   LPSTR  lpszRetBuf )
{
  SOCKET              conn_socket;
  struct sockaddr_in  server;
  int                 ret;
  char                szBuf[256],szPBuf[256];
  struct hostent      *hp;

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("SendCommand: enter.");
#endif

  conn_socket = socket( AF_INET, SOCK_STREAM, 0 );
  if ( conn_socket == INVALID_SOCKET ){
    sprintf(szPBuf,"SendCommand: Error: socket() fail. (%d)",WSAGetLastError());
    OutputDebugLog(szPBuf);
    return ( FALSE );
  }

  hp = gethostbyname("localhost");
  memset(&server,0,sizeof(server));
  memcpy(&(server.sin_addr),hp->h_addr,hp->h_length);
  server.sin_family = AF_INET;
  server.sin_port = htons(port_num);
  ret = connect( conn_socket, (struct sockaddr*)&server, sizeof(server) );
  if ( ret == SOCKET_ERROR ){
    sprintf(szPBuf,"SendCommand: Error: connect() fail. (%d)",WSAGetLastError());
    OutputDebugLog(szPBuf);
    goto Error;
  }

  ret = recv(conn_socket,szBuf,sizeof(szBuf),0);
  if ( ret == SOCKET_ERROR ){
    sprintf(szPBuf,"SendCommand: Error: recv() fail. (%d)",WSAGetLastError());
    OutputDebugLog(szPBuf);
    goto Error;
  }
  ret = send(conn_socket,lpszSendCommand,strlen(lpszSendCommand),0);
  if ( ret == SOCKET_ERROR ){
    sprintf(szPBuf,"SendCommand: Error: send() fail. (%d)",WSAGetLastError());
    OutputDebugLog(szPBuf);
    goto Error;
  }
  ret = recv(conn_socket,szBuf,sizeof(szBuf),0);
  if ( ret == SOCKET_ERROR ){
    sprintf(szPBuf,"SendCommand: Error: recv() fail. (%d)",WSAGetLastError());
    OutputDebugLog(szPBuf);
    goto Error;
  }

  /* close socket */
  ret = closesocket(conn_socket);
  if ( ret != 0 )
    goto Error;

  OutputDebugLog("SendCommand: server closed connection.");
  
  strcpy(lpszRetBuf,szBuf);

#ifdef DETAIL_ERROR_MESSAGE
  OutputDebugLog("SendCommand: leave.");
#endif

  return ( TRUE );

 Error:
  closesocket( conn_socket );
  return ( FALSE );

}
